// Gaia Sky port of shadertoy shader "Helix Nebula" by izutionix.
// Licensed under CC-BY-NC-SA.
// Source: https://www.shadertoy.com/view/cdK3Wy
#version 330 core

// Uniforms which are always available
uniform mat4 u_worldTrans;
uniform vec3 u_cameraPos;
uniform float u_bodySize;
uniform vec2 u_cameraNearFar;
uniform float u_cameraK;
uniform float u_time;
uniform float u_simuTime;
// Unit conversion.
uniform float u_kmToU;

#ifdef ssrFlag
#include <shader/lib/ssr.frag.glsl>
#endif // ssrFlag

// INPUT
struct VertexData {
    vec2 texCoords;
    vec3 normal;
    vec3 viewDir;
    vec3 ambientLight;
    float opacity;
    vec4 color;
    vec3 fragPosWorld;
    mat3 tbn;
};
in VertexData v_data;

// OUTPUT
layout(location = 0) out vec4 fragColor;
layout(location = 1) out vec4 layerBuffer;

#include <shader/lib/logdepthbuff.glsl>
#include <$data/catalog-nebulae/shader/lib/nebulae-lib.glsl>

#define DIST_SCALE (u_kmToU * 1.0e6)
#define FACTOR (1.0 / DIST_SCALE * 2.0e-7)

#define ITERATIONS 70.0

float Disk(vec3 p, vec3 t) {
    vec2 q = vec2(length2(p.xy) - t.x, p.z * 0.7);
    return max(length8(q) - t.y, abs(p.z) - t.z);
}

//==============================================================
// otaviogood's noise from https://www.shadertoy.com/view/ld2SzK
//--------------------------------------------------------------
// This spiral noise works by successively adding and rotating sin waves while increasing frequency.
// It should work the same on all computers since it's not based on a hash function like some other noises.
// It can be much faster than other noise functions if you're ok with some repetition.
const float nudge = 0.9; // size of perpendicular vector
float normalizer = 1.0 / sqrt(1.0 + nudge * nudge); // pythagorean theorem on that perpendicular to maintain scale
float SpiralNoiseC(vec3 p) {
    float n = 0.0; // noise amount
    float iter = 2.0;
    for (int i = 0; i < 8; i++) {
        // add sin and cos scaled inverse with the frequency
        n += -abs(sin(p.y * iter) + cos(p.x * iter)) / iter; // abs for a ridged look
        // rotate by adding perpendicular and scaling down
        p.xy += vec2(p.y, -p.x) * nudge;
        p.xy *= normalizer;
        // rotate on other axis
        p.xz += vec2(p.z, -p.x) * nudge;
        p.xz *= normalizer;
        // increase the frequency
        iter *= 1.733733;
    }
    return n;
}

float NebulaNoise(vec3 p) {
    float final = Disk(p.xzy, vec3(2.0, 2.0, 0.3));
    final += fbm(p * 60.0);
    final += SpiralNoiseC(p.zxy * 0.7 + 1.6) * 2.0;

    return final;
}

float map(vec3 p) {
    p = p.yzx;
    vec4 q = get_quat_rotation(vec3(1.0, 1.0, 1.0), -PI / 3.0);
    p = rotate(p, q);
    float NebNoise = abs(NebulaNoise(p / 0.5) * 0.5);

    return NebNoise + 0.07;
}
//--------------------------------------------------------------

// assign color to the media
vec3 computeColor(float density, float radius) {
    // color based on density alone, gives impression of occlusion within
    // the media
    //vec3 result = mix( vec3(1.0,0.9,0.8), vec3(0.4,0.15,0.1), density );
    vec3 result = mix(vec3(1.0), vec3(0.5), density);

    // color added to the media
    vec3 colCenter = 7.0 * vec3(0.8, 0.9, 1.0).rgb;
    vec3 colEdge = 1.5 * vec3(0.48, 0.53, 0.5).rgb;
    result *= mix(colCenter, colEdge, min((radius + 0.05) / 0.9, 1.15));

    return result;
}

vec4 raymarch(vec3 camPos, vec3 rayDir, vec3 objPos, vec2 dists, out float s) {
    // Ray direction.
    vec3 rd = rayDir;
    // Ray origin.
    vec3 ro = camPos;

    // ld, td: local, total density
    // w: weighting factor
    float ld = 0.0, td = 0.0, w = 0.0;

    // d: distance function
    float d = 1.0;

    const float h = 0.1;

    vec4 sum = vec4(0.0);

    float min_dist = dists.x, max_dist = dists.y;

    if (min_dist >= 0.0 && max_dist >= 0.0) {
        // t: length of the ray
        float t = min_dist;

        // raymarch loop
        for (int i = 0; i < ITERATIONS; i++) {
            vec3 pos = ro + t * rd;

            // Loop break conditions.
            if (td > 0.9 || d < 0.02 * t || t > 30.0 || sum.a > 0.99) break;

            // evaluate distance function
            float d = map(pos);

            // change this string to control density
            d = max(d, 0.0);

            // point light calculations
            vec3 ldst = vec3(0.0) - pos;
            float lDist = max(length(ldst), 0.001);

            // the color of light
            float _T = lDist * 2.3 + 2.6; // <-v endless tweaking
            vec3 lightColor = 0.4 + 0.5 * cos(_T + PI * 0.5 * vec3(-0.5, 0.05, 0.5)); //vec3(1.0,0.5,0.25);

            // Star
            sum.rgb += (vec3(0.57, 1.85, 1.00) / (lDist * lDist * 10.0) / 70.0);
            // Bloom
            sum.rgb += (lightColor / exp(lDist * lDist * lDist * 0.05) / 50.0);

            if (d < h) {
                // compute local density
                ld = h - d;

                // compute weighting factor
                w = (1.0 - td) * ld;

                // accumulate density
                td += w + 1.0 / 200.0;

                vec4 col = vec4(computeColor(td, lDist), td);

                // emission
                sum += sum.a * vec4(sum.rgb, 0.0) * 0.2;

                // uniform scale density
                col.a *= 0.2;
                // colour by alpha
                col.rgb *= col.a;
                // alpha blend in contribution
                sum = sum + col * (1.0 - sum.a);
            }
            td += 1.0 / 60.0;
            // Optimize step size near the camera
            t += max(d * 0.1 * max(min(length(ldst), length(ro)), 1.0), 0.01);
        }
        td *= smoothstep(0.0, 0.8, td);

        // simple scattering
        sum *= 1.0 / exp(ld * 0.2) * 0.6;
        sum = clamp(sum, 0.0, 1.0);
        sum.xyz = sum.xyz * sum.xyz * (3.0 - 2.0 * sum.xyz);
    }
    return vec4(sum.xyz, td);
}

void main() {
    // Directions and positions.
    // Object is alwyas at origin in our reference frame.
    vec3 camPos = u_cameraPos * FACTOR;
    vec3 rayDir = normalize(v_data.fragPosWorld);
    vec3 objPos = vec3(0.0, 0.0, 0.0);
    float radius = u_bodySize;

    vec2 dists = sphereIntersectionDistances(camPos, rayDir, objPos, radius);

    // Ray marching
    float d = 0.0;
    vec4 rmcol = raymarch(camPos, rayDir, objPos, dists, d);
    fragColor = rmcol;

    // Render outline
    // vec2 uv = v_data.texCoords;
    // if (uv.x < 0.005 || uv.x > 0.995 || uv.y < 0.005 || uv.y > 0.995) {
    //     d = length(v_data.fragPosWorld);
    //     fragColor = vec4(1.0, 0.0, 0.0, 1.0);
    // }

    // Logarithmic depth buffer.
    if (d > 0.0) {
        gl_FragDepth = getDepthValue(d, u_cameraNearFar.y, u_cameraK);
    }

    #ifdef ssrFlag
    ssrBuffers();
    #endif // ssrFlag
}
